from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytest
from caraml.upi.v1 import table_pb2, type_pb2, upi_pb2
from merlin.model import (
    PYFUNC_EXTRA_ARGS_KEY,
    PYFUNC_MODEL_INPUT_KEY,
    PyFuncModel,
    PyFuncV3Model,
)
from merlin.protocol import Protocol
from merlin.pyfunc import (
    PYFUNC_GRPC_CONTEXT,
    PYFUNC_PROTOCOL_KEY,
    ModelInput,
    ModelOutput,
    PyFuncOutput,
    Values,
)
from mlflow.pyfunc import PythonModelContext

request = {"instances": [[1, 2, 3, 4]]}

response = {"predictions": [[1]]}

headers = {"key": "value"}

upiv1_request = upi_pb2.PredictValuesRequest()
upiv1_response = upi_pb2.PredictValuesResponse()
grpc_context = {}

pyfunc_model_input = ModelInput(
    features=Values(
        columns=["featureA", "featureB", "featureC"],
        data=[[0.1, 0.2, "debit"], [0.12, 0.4, "cc"], [0.42, 0.2, "debit"]],
    ),
    entities=Values(
        columns=["order_id", "customer_id"],
        data=[["order1", "111"], ["order1", "112"], ["order1", "113"]],
    ),
    prediction_ids=["prediction1", "prediction2", "prediction3"],
)
pyfunc_model_output = ModelOutput(
    predictions=Values(
        columns=["prediction_score", "prediction_label"],
        data=[[0.95, "complete"], [0.43, "incomplete"], [0.59, "complete"]],
    ),
    prediction_ids=["prediction1", "prediction2", "prediction3"],
)


def test_http():
    pyfunc_model = PyFuncModel()
    pyfunc_model.infer = MagicMock(return_value=response)

    context = PythonModelContext(artifacts={})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == response
    pyfunc_model.infer.assert_called_with(request)


def test_http_headers():
    pyfunc_model = PyFuncModel()
    pyfunc_model.infer = MagicMock(return_value=response)

    context = PythonModelContext(artifacts={})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
        PYFUNC_EXTRA_ARGS_KEY: {"headers": headers},
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == response
    pyfunc_model.infer.assert_called_with(request, headers=headers)


def test_explicit_protocol():
    pyfunc_model = PyFuncModel()
    pyfunc_model.infer = MagicMock(return_value=response)

    context = PythonModelContext(artifacts={})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
        PYFUNC_PROTOCOL_KEY: Protocol.HTTP_JSON,
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == response
    pyfunc_model.infer.assert_called_with(request)


def test_upiv1():
    pyfunc_model = PyFuncModel()
    pyfunc_model.upiv1_infer = MagicMock(return_value=upiv1_response)

    context = PythonModelContext(artifacts={})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_PROTOCOL_KEY: Protocol.UPI_V1,
        PYFUNC_MODEL_INPUT_KEY: upiv1_request,
        PYFUNC_GRPC_CONTEXT: grpc_context,
    }

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == upiv1_response
    pyfunc_model.upiv1_infer.assert_called_with(upiv1_request, grpc_context)


def test_pyfuncv3_rest():
    pyfunc_model = PyFuncV3Model()
    http_response = {
        "columns": ["prediction_score", "prediction_label"],
        "data": [[0.95, "complete"], [0.43, "incomplete"], [0.59, "complete"]],
    }
    pyfunc_model.preprocess = MagicMock(return_value=pyfunc_model_input)
    pyfunc_model.infer = MagicMock(return_value=pyfunc_model_output)
    pyfunc_model.postprocess = MagicMock(return_value=http_response)
    context = PythonModelContext(artifacts={})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_MODEL_INPUT_KEY: request,
    }

    exp_response = PyFuncOutput(
        http_response=http_response,
        model_input=pyfunc_model_input,
        model_output=pyfunc_model_output,
    )

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == exp_response
    pyfunc_model.preprocess.assert_called_with(model_input[PYFUNC_MODEL_INPUT_KEY])
    pyfunc_model.infer.assert_called_with(pyfunc_model_input)
    pyfunc_model.postprocess.assert_called_with(
        pyfunc_model_output, model_input[PYFUNC_MODEL_INPUT_KEY]
    )


def test_pyfuncv3_upi():
    pyfunc_model = PyFuncV3Model()
    upi_response = upi_pb2.PredictValuesResponse(
        prediction_result_table=table_pb2.Table(
            name="prediction_result_table",
            columns=[
                table_pb2.Column(name="prediction_score", type=type_pb2.TYPE_DOUBLE),
                table_pb2.Column(name="prediction_label", type=type_pb2.TYPE_STRING),
            ],
            rows=[
                table_pb2.Row(
                    row_id="0",
                    values=[
                        table_pb2.Value(double_value=0.95),
                        table_pb2.Value(string_value="complete"),
                    ],
                ),
                table_pb2.Row(
                    row_id="1",
                    values=[
                        table_pb2.Value(double_value=0.43),
                        table_pb2.Value(string_value="incomplete"),
                    ],
                ),
                table_pb2.Row(
                    row_id="2",
                    values=[
                        table_pb2.Value(double_value=0.59),
                        table_pb2.Value(string_value="complete"),
                    ],
                ),
            ],
        )
    )
    pyfunc_model.upiv1_preprocess = MagicMock(return_value=pyfunc_model_input)
    pyfunc_model.infer = MagicMock(return_value=pyfunc_model_output)
    pyfunc_model.upiv1_postprocess = MagicMock(return_value=upi_response)
    context = PythonModelContext(artifacts={})

    pyfunc_model.load_context(context)
    model_input = {
        PYFUNC_PROTOCOL_KEY: Protocol.UPI_V1,
        PYFUNC_MODEL_INPUT_KEY: upiv1_request,
        PYFUNC_GRPC_CONTEXT: grpc_context,
    }

    exp_response = PyFuncOutput(
        upi_response=upi_response,
        model_input=pyfunc_model_input,
        model_output=pyfunc_model_output,
    )

    actual_resp = pyfunc_model.predict(context=context, model_input=model_input)
    assert actual_resp == exp_response
    pyfunc_model.upiv1_preprocess.assert_called_with(
        model_input[PYFUNC_MODEL_INPUT_KEY], grpc_context
    )
    pyfunc_model.infer.assert_called_with(pyfunc_model_input)
    pyfunc_model.upiv1_postprocess.assert_called_with(
        pyfunc_model_output, model_input[PYFUNC_MODEL_INPUT_KEY]
    )


@pytest.mark.parametrize(
    "features,row_ids,use_deprecated_prediction_ids,expected",
    [
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=[[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
            ),
            ["prediction_1", "prediction_2", "prediction3"],
            False,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2", "prediction3"],
            },
        ),
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=np.array([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]]),
            ),
            ["prediction_1", "prediction_2", "prediction3"],
            False,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2", "prediction3"],
            },
        ),
        (
            pd.DataFrame(
                [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], columns=["col1", "col2", "col3"]
            ),
            ["prediction_1", "prediction_2", "prediction3"],
            False,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2", "prediction3"],
            },
        ),
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=[[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
            ),
            ["prediction_1", "prediction_2", "prediction3"],
            True,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2", "prediction3"],
            },
        ),
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=np.array([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]]),
            ),
            ["prediction_1", "prediction_2", "prediction3"],
            True,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2", "prediction3"],
            },
        ),
        (
            pd.DataFrame(
                [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], columns=["col1", "col2", "col3"]
            ),
            ["prediction_1", "prediction_2", "prediction3"],
            True,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2", "prediction3"],
            },
        ),
    ],
)
def test_model_input_features(
    features, row_ids, use_deprecated_prediction_ids, expected
):
    model_input = ModelInput(features=features, row_ids=row_ids)
    if use_deprecated_prediction_ids:
        model_input = ModelInput(features=features, prediction_ids=row_ids)
    assert model_input.features_dict() == expected


@pytest.mark.parametrize(
    "entities,row_ids,use_deprecated_prediction_ids,expected",
    [
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=[[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
            ),
            ["prediction_1", "prediction_2"],
            False,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=np.array([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]]),
            ),
            ["prediction_1", "prediction_2"],
            False,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            pd.DataFrame(
                [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], columns=["col1", "col2", "col3"]
            ),
            ["prediction_1", "prediction_2"],
            False,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (None, None, False, None),
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=[[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
            ),
            ["prediction_1", "prediction_2"],
            True,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            Values(
                columns=["col1", "col2", "col3"],
                data=np.array([[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]]),
            ),
            ["prediction_1", "prediction_2"],
            True,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            pd.DataFrame(
                [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]], columns=["col1", "col2", "col3"]
            ),
            ["prediction_1", "prediction_2"],
            True,
            {
                "columns": ["col1", "col2", "col3"],
                "data": [[0.1, 0.2, 0.3], [0.2, 0.25, 0.3]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (None, None, True, None),
    ],
)
def test_model_input_entities(
    entities, row_ids, use_deprecated_prediction_ids, expected
):
    model_input = ModelInput(
        features=Values(columns=["col1"], data=[[0.2], [0.3]]),
        entities=entities,
        row_ids=row_ids,
    )
    if use_deprecated_prediction_ids:
        model_input = ModelInput(
            features=Values(columns=["col1"], data=[[0.2], [0.3]]),
            entities=entities,
            prediction_ids=row_ids,
        )
    assert model_input.entities_dict() == expected


@pytest.mark.parametrize(
    "predictions,row_ids,use_deprecated_prediction_ids,expected",
    [
        (
            Values(columns=["prediction_score"], data=[[0.1], [0.2]]),
            ["prediction_1", "prediction_2"],
            False,
            {
                "columns": ["prediction_score"],
                "data": [[0.1], [0.2]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            Values(columns=["prediction_score"], data=np.array([[0.1], [0.2]])),
            ["prediction_1", "prediction_2"],
            False,
            {
                "columns": ["prediction_score"],
                "data": [[0.1], [0.2]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            pd.DataFrame([[0.1], [0.2]], columns=["prediction_score"]),
            ["prediction_1", "prediction_2"],
            False,
            {
                "columns": ["prediction_score"],
                "data": [[0.1], [0.2]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (None, None, False, None),
        (
            Values(columns=["prediction_score"], data=[[0.1], [0.2]]),
            ["prediction_1", "prediction_2"],
            True,
            {
                "columns": ["prediction_score"],
                "data": [[0.1], [0.2]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            Values(columns=["prediction_score"], data=np.array([[0.1], [0.2]])),
            ["prediction_1", "prediction_2"],
            True,
            {
                "columns": ["prediction_score"],
                "data": [[0.1], [0.2]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (
            pd.DataFrame([[0.1], [0.2]], columns=["prediction_score"]),
            ["prediction_1", "prediction_2"],
            True,
            {
                "columns": ["prediction_score"],
                "data": [[0.1], [0.2]],
                "row_ids": ["prediction_1", "prediction_2"],
            },
        ),
        (None, None, True, None),
    ],
)
def test_model_output_predictions(
    predictions, row_ids, use_deprecated_prediction_ids, expected
):
    model_output = ModelOutput(predictions=predictions, row_ids=row_ids)
    if use_deprecated_prediction_ids:
        model_output = ModelOutput(predictions=predictions, prediction_ids=row_ids)
    assert model_output.predictions_dict() == expected



import pytest
import client
from merlin.model_observability import ModelObservability, GroundTruthSource, GroundTruthJob, PredictionLogIngestionResourceRequest

@pytest.mark.unit
@pytest.mark.parametrize(
    "response,expected", [
        (
            client.ModelObservability(
                enabled=True,
                ground_truth_source=client.GroundTruthSource(
                    table_urn="table_urn",
                    event_timestamp_column="event_timestamp_column",
                    source_project="dwh_project"
                ),
                ground_truth_job=client.GroundTruthJob(
                    cron_schedule="cron_schedule",
                        service_account_secret_name="service_account_secret_name",
                        start_day_offset_from_now=1,
                        end_day_offset_from_now=1,
                        cpu_request="cpu_request",
                        cpu_limit="cpu_limit",
                        memory_request="memory_request",
                        memory_limit="memory_limit",
                        grace_period_day=1
                ),
                prediction_log_ingestion_resource_request=client.PredictionLogIngestionResourceRequest(
                    replica=1,
                    cpu_request="1",
                    memory_request="1Gi"
                )
            ),
            ModelObservability(
                enabled=True,
                ground_truth_source=GroundTruthSource(
                    table_urn="table_urn",
                    event_timestamp_column="event_timestamp_column",
                    source_project="dwh_project"
                ),
                ground_truth_job=GroundTruthJob(
                    cron_schedule="cron_schedule",
                    service_account_secret_name="service_account_secret_name",
                    start_day_offset_from_now=1,
                    end_day_offset_from_now=1,
                    cpu_request="cpu_request",
                    cpu_limit="cpu_limit",

                    memory_request="memory_request",
                    memory_limit="memory_limit",
                    grace_period_day=1
                ),
                prediction_log_ingestion_resource_request=PredictionLogIngestionResourceRequest(
                    replica=1,
                    cpu_request="1",
                    memory_request="1Gi"
                )
            )
        )
    ]
)
def test_from_model_observability_response(response, expected):
    assert ModelObservability.from_model_observability_response(response) == expected
    assert expected.to_model_observability_request() == response


# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
from inspect import getfullargspec
import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional
from pydantic import BaseModel, Field, StrictStr, ValidationError, field_validator
from client.models.binary_classification_output import BinaryClassificationOutput
from client.models.ranking_output import RankingOutput
from client.models.regression_output import RegressionOutput
from typing import Union, Any, List, TYPE_CHECKING, Optional, Dict
from typing_extensions import Literal
from pydantic import StrictStr, Field
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

MODELPREDICTIONOUTPUT_ONE_OF_SCHEMAS = ["BinaryClassificationOutput", "RankingOutput", "RegressionOutput"]

class ModelPredictionOutput(BaseModel):
    """
    ModelPredictionOutput
    """
    # data type: BinaryClassificationOutput
    oneof_schema_1_validator: Optional[BinaryClassificationOutput] = None
    # data type: RankingOutput
    oneof_schema_2_validator: Optional[RankingOutput] = None
    # data type: RegressionOutput
    oneof_schema_3_validator: Optional[RegressionOutput] = None
    actual_instance: Optional[Union[BinaryClassificationOutput, RankingOutput, RegressionOutput]] = None
    one_of_schemas: List[str] = Literal["BinaryClassificationOutput", "RankingOutput", "RegressionOutput"]

    model_config = {
        "validate_assignment": True
    }


    discriminator_value_class_map: Dict[str, str] = {
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_oneof(cls, v):
        instance = ModelPredictionOutput.model_construct()
        error_messages = []
        match = 0
        # validate data type: BinaryClassificationOutput
        if not isinstance(v, BinaryClassificationOutput):
            error_messages.append(f"Error! Input type `{type(v)}` is not `BinaryClassificationOutput`")
        else:
            match += 1
        # validate data type: RankingOutput
        if not isinstance(v, RankingOutput):
            error_messages.append(f"Error! Input type `{type(v)}` is not `RankingOutput`")
        else:
            match += 1
        # validate data type: RegressionOutput
        if not isinstance(v, RegressionOutput):
            error_messages.append(f"Error! Input type `{type(v)}` is not `RegressionOutput`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when setting `actual_instance` in ModelPredictionOutput with oneOf schemas: BinaryClassificationOutput, RankingOutput, RegressionOutput. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when setting `actual_instance` in ModelPredictionOutput with oneOf schemas: BinaryClassificationOutput, RankingOutput, RegressionOutput. Details: " + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # use oneOf discriminator to lookup the data type
        _data_type = json.loads(json_str).get("output_class")
        if not _data_type:
            raise ValueError("Failed to lookup data type from the field `output_class` in the input.")

        # check if data type is `BinaryClassificationOutput`
        if _data_type == "BinaryClassificationOutput":
            instance.actual_instance = BinaryClassificationOutput.from_json(json_str)
            return instance

        # check if data type is `RankingOutput`
        if _data_type == "RankingOutput":
            instance.actual_instance = RankingOutput.from_json(json_str)
            return instance

        # check if data type is `RegressionOutput`
        if _data_type == "RegressionOutput":
            instance.actual_instance = RegressionOutput.from_json(json_str)
            return instance

        # deserialize data into BinaryClassificationOutput
        try:
            instance.actual_instance = BinaryClassificationOutput.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into RankingOutput
        try:
            instance.actual_instance = RankingOutput.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into RegressionOutput
        try:
            instance.actual_instance = RegressionOutput.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when deserializing the JSON string into ModelPredictionOutput with oneOf schemas: BinaryClassificationOutput, RankingOutput, RegressionOutput. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when deserializing the JSON string into ModelPredictionOutput with oneOf schemas: BinaryClassificationOutput, RankingOutput, RegressionOutput. Details: " + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        to_json = getattr(self.actual_instance, "to_json", None)
        if callable(to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> Dict:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        to_dict = getattr(self.actual_instance, "to_dict", None)
        if callable(to_dict):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())


